package eu.jonahbauer.raytracing.render.renderer.pdf;

import eu.jonahbauer.raytracing.math.Vec3;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;
import java.util.random.RandomGenerator;

/**
 * Mixes between two probability density functions (pdf) using a weight. When the weight is closer to zero, the
 * influence of the second pdf is stronger. When the weight is closer to one, the influence of the first pdf is stronger.
 * @param a the first probability density function
 * @param b the second probability density function
 * @param weight a weight in the range [0, 1]
 */
public record MixtureProbabilityDensityFunction(
        @NotNull ProbabilityDensityFunction a,
        @NotNull ProbabilityDensityFunction b,
        double weight
) implements ProbabilityDensityFunction {
    public MixtureProbabilityDensityFunction(@NotNull ProbabilityDensityFunction a, @NotNull ProbabilityDensityFunction b) {
        this(a, b, 0.5);
    }

    public MixtureProbabilityDensityFunction {
        Objects.requireNonNull(a);
        Objects.requireNonNull(b);
        weight = Math.clamp(weight, 0, 1);
    }

    @Override
    public double value(@NotNull Vec3 direction) {
        var v = a.value(direction);
        var w = b.value(direction);
        return Math.fma(weight, v, Math.fma(-weight, w, w));
    }

    @Override
    public @NotNull Vec3 generate(@NotNull RandomGenerator random) {
        if (random.nextDouble() < weight) {
            return a.generate(random);
        } else {
            return b.generate(random);
        }
    }
}
